use crate::datastore::{SigningError, Store};
use std::sync::Arc;

use salvo::affix;
use salvo::prelude::*;

enum SignError {
    NoPubkey,
    NoMessage,
    Store(SigningError),
}
#[async_trait]
impl Writer for SignError {
    async fn write(self, _req: &mut Request, _depot: &mut Depot, res: &mut Response) {
        res.status_code(StatusCode::BAD_REQUEST);
        match self {
            SignError::NoPubkey => res.render("pubkey must be specified as GET param"),
            SignError::NoMessage => res.render("something must be specified as GET param"),
            SignError::Store(e) => res.render(format!("{e}")),
        };
    }
}

#[handler]
async fn homepage(depot: &mut Depot) -> String {
    let ds = depot.obtain::<Arc<Store>>().unwrap();
    ds.tabled_node_list().await // TODO: make this paginated
}

#[handler]
async fn sign(req: &mut Request, depot: &mut Depot) -> Result<String, SignError> {
    let ds = depot.obtain::<Arc<Store>>().unwrap();
    let pubkey = match req.query::<String>("pubkey") {
        Some(k) => k,
        None => return Err(SignError::NoPubkey),
    };

    let something = match req.query::<String>("something") {
        Some(k) => k,
        None => return Err(SignError::NoMessage),
    };

    match ds.sign_message_with_key(&something, &pubkey).await {
        Ok(s) => Ok(s),
        Err(e) => Err(SignError::Store(e)),
    }
}

pub async fn init(ds: Arc<Store>) {
    let acceptor = TcpListener::new("0.0.0.0:31372").bind().await;
    let router = Router::new()
        .hoop(affix::inject(ds))
        .get(homepage)
        .push(Router::with_path("sign").get(sign));
    println!("{:?}", router);
    Server::new(acceptor).serve(router).await;
}
