#![allow(dead_code)]
use ed25519_dalek::SigningKey;
use ed25519_dalek::KEYPAIR_LENGTH;
use std::net::AddrParseError;
use std::net::Ipv4Addr;
use std::time::{Duration, SystemTime, UNIX_EPOCH};
use tokio::fs::File;
use tokio::io::{AsyncReadExt, AsyncSeekExt, AsyncWriteExt, SeekFrom};
use tokio::sync::Mutex;

const DATA_SIZE: usize = 76;

enum Error {
    CorruptedIP,
}

impl From<AddrParseError> for Error {
    fn from(_: AddrParseError) -> Self {
        Error::CorruptedIP
    }
}

#[derive(Clone)]
struct Node {
    ip: Ipv4Addr,
    keypair: SigningKey,
    joined_at: SystemTime,
}

impl TryFrom<(&str, SigningKey, SystemTime)> for Node {
    type Error = Error;
    fn try_from(value: (&str, SigningKey, SystemTime)) -> Result<Self, Self::Error> {
        Ok(Self {
            ip: value.0.parse()?,
            keypair: value.1,
            joined_at: value.2,
        })
    }
}

impl Node {
    fn ip_as_string(&self) -> String {
        self.ip.to_string()
    }

    fn signing_key(&self) -> SigningKey {
        self.keypair.clone()
    }

    fn to_bytes(self) -> [u8; DATA_SIZE] {
        let mut result = [0; DATA_SIZE];
        result[0..4].copy_from_slice(&self.ip.octets());
        result[4..68].copy_from_slice(&self.keypair.to_keypair_bytes());
        result[68..DATA_SIZE].copy_from_slice(
            &self
                .joined_at
                .duration_since(UNIX_EPOCH)
                .unwrap()
                .as_secs()
                .to_le_bytes(),
        );
        result
    }

    fn from_bytes(bytes: [u8; DATA_SIZE]) -> Self {
        let ip: [u8; 4] = bytes[0..4].try_into().unwrap();
        let ip: Ipv4Addr = ip.into();
        let keypair: [u8; KEYPAIR_LENGTH] = bytes[4..68].try_into().unwrap();
        let keypair: SigningKey = SigningKey::from_keypair_bytes(&keypair).unwrap();
        let joined_at: [u8; 8] = bytes[68..DATA_SIZE].try_into().unwrap();
        let joined_at: u64 = u64::from_le_bytes(joined_at);
        let joined_at = SystemTime::UNIX_EPOCH + Duration::from_secs(joined_at);
        Self {
            ip,
            keypair,
            joined_at,
        }
    }
}

struct FileManager {
    file: Mutex<File>,
}

impl FileManager {
    async fn init(path: &str) -> std::io::Result<Self> {
        let file = File::open(path).await?;
        Ok(Self {
            file: Mutex::new(file),
        })
    }

    async fn append_node(&self, node: Node) -> std::io::Result<()> {
        let mut file = self.file.lock().await;
        file.seek(SeekFrom::End(0)).await?;
        file.write_all(&node.to_bytes()).await?;
        file.flush().await?;
        Ok(())
    }

    async fn get_node_by_id(&self, id: u64) -> std::io::Result<Node> {
        let mut file = self.file.lock().await;
        file.seek(SeekFrom::Start(
            id.wrapping_mul(DATA_SIZE.try_into().unwrap_or(0)),
        ))
        .await?;
        let mut node_bytes = [0; DATA_SIZE];
        file.read_exact(&mut node_bytes).await?;
        Ok(Node::from_bytes(node_bytes))
    }
}

#[cfg(test)]
mod tests {

    use super::*;
    use ed25519_dalek::SigningKey;
    use rand::rngs::OsRng;
    use std::io::Result;
    use tokio::fs::remove_file;
    use tokio::io::AsyncWriteExt;

    const TEST_FILE_PREFIX: &str = ".tmp/test_";
    fn get_test_file_name(function: &str) -> String {
        TEST_FILE_PREFIX.to_string() + function
    }
    async fn setup_test_file(function: &str) -> Result<FileManager> {
        let path = get_test_file_name(function);
        let mut file = File::create(path.clone()).await?;
        file.flush().await?;
        drop(file);
        FileManager::init(&path).await
    }

    #[test]
    fn test_node_to_bytes_and_back() {
        let keypair = SigningKey::generate(&mut OsRng);

        let original_node = Node {
            ip: "192.168.1.1".parse().unwrap(),
            keypair: keypair.clone(),
            joined_at: SystemTime::now(),
        };

        let node_bytes = original_node.clone().to_bytes();
        let restored_node = Node::from_bytes(node_bytes);

        assert_eq!(original_node.ip_as_string(), restored_node.ip_as_string());
        assert_eq!(
            original_node.keypair.to_keypair_bytes(),
            restored_node.keypair.to_keypair_bytes()
        );
        assert_eq!(
            original_node
                .joined_at
                .duration_since(UNIX_EPOCH)
                .unwrap()
                .as_secs(),
            restored_node
                .joined_at
                .duration_since(UNIX_EPOCH)
                .unwrap()
                .as_secs()
        );
    }

    #[tokio::test]
    async fn setup_file_manager() {
        match setup_test_file("setup_file_manager").await {
            Err(e) => {
                panic!("Could not init File Manager: {}", e);
            }
            _ => {}
        }
    }

    #[tokio::test]
    async fn test_file_manager_append_and_retrieve_node() -> Result<()> {
        let function_name = "test_file_manager_append_and_retrieve_node";
        let manager = setup_test_file(function_name).await?;

        let keypair = SigningKey::generate(&mut OsRng);

        let node = Node {
            ip: "192.168.1.1".parse().unwrap(),
            keypair: keypair.clone(),
            joined_at: SystemTime::now(),
        };

        manager.append_node(node.clone()).await.unwrap();

        let retrieved_node = manager.get_node_by_id(0).await.unwrap();

        assert_eq!(node.ip_as_string(), retrieved_node.ip_as_string());
        assert_eq!(
            node.keypair.to_keypair_bytes(),
            retrieved_node.keypair.to_keypair_bytes()
        );
        assert_eq!(node.joined_at, retrieved_node.joined_at);

        remove_file(get_test_file_name(function_name)).await?;

        Ok(())
    }

    #[tokio::test]
    async fn test_file_manager_multiple_nodes() -> Result<()> {
        let function_name = "test_file_manager_multiple_nodes";
        let manager = setup_test_file(function_name).await?;

        let keypair1 = SigningKey::generate(&mut OsRng);
        let node1 = Node {
            ip: "192.168.1.1".parse().unwrap(),
            keypair: keypair1.clone(),
            joined_at: SystemTime::now(),
        };

        let keypair2 = SigningKey::generate(&mut OsRng);
        let node2 = Node {
            ip: "10.0.0.1".parse().unwrap(),
            keypair: keypair2.clone(),
            joined_at: SystemTime::now(),
        };

        manager.append_node(node1.clone()).await.unwrap();
        manager.append_node(node2.clone()).await.unwrap();

        let retrieved_node1 = manager.get_node_by_id(0).await?;
        let retrieved_node2 = manager.get_node_by_id(1).await?;

        assert_eq!(node1.ip_as_string(), retrieved_node1.ip_as_string());
        assert_eq!(
            node1.keypair.to_keypair_bytes(),
            retrieved_node1.keypair.to_keypair_bytes()
        );

        assert_eq!(node2.ip_as_string(), retrieved_node2.ip_as_string());
        assert_eq!(
            node2.keypair.to_keypair_bytes(),
            retrieved_node2.keypair.to_keypair_bytes()
        );

        remove_file(get_test_file_name(function_name)).await?;

        Ok(())
    }
}
